@@ -110,7 +110,10 @@ namespace Tomboy.Sync
 	/// <summary>
 	/// Handle a note conflict
 	/// </summary>
-	public delegate void NoteConflictHandler (NoteManager manager, Note localConflictNote, NoteUpdate remoteNote);
+	public delegate void NoteConflictHandler (NoteManager manager,
+	                                          Note localConflictNote,
+	                                          NoteUpdate remoteNote,
+	                                          IList<string> noteUpdateTitles);
 	
 	public class SyncManager
 	{
@@ -211,8 +214,10 @@ namespace Tomboy.Sync
 		public static void SynchronizationThread ()
 		{
 			// TODO: Try/finally this entire method so GUI doesn't hang?
+			SyncServiceAddin addin = null;
+			try {
 
-			SyncServiceAddin addin = GetConfiguredSyncService ();
+			addin = GetConfiguredSyncService ();
 			if (addin == null) {
 				SetState (SyncState.NoConfiguredSyncService);
 				Logger.Debug ("GetConfiguredSyncService is null");
@@ -274,6 +279,13 @@ Logger.Debug ("8");
 				server.GetNoteUpdatesSince (client.LastSynchronizedRevision);
 			Logger.Debug ("Sync: " + noteUpdates.Count + " updates since rev " + client.LastSynchronizedRevision.ToString ());
 
+					// Gather list of new/updated note titles
+					// for title conflict handling purposes.
+					List<string> noteUpdateTitles = new List<string> ();
+					foreach (NoteUpdate noteUpdate in noteUpdates.Values)
+						if (!string.IsNullOrEmpty (noteUpdate.Title))
+							noteUpdateTitles.Add (noteUpdate.Title);
+					
 			// First, check for new local notes that might have title conflicts
 			// with the updates coming from the server.  Prompt the user if necessary.
 			// TODO: Lots of searching here and in the next foreach...
@@ -285,7 +297,7 @@ Logger.Debug ("8");
 					Note existingNote = NoteMgr.Find (noteUpdate.Title);
 					if (existingNote != null) {
 						if (NoteConflictDetected != null) {
-							NoteConflictDetected (NoteMgr, existingNote, noteUpdate);
+							NoteConflictDetected (NoteMgr, existingNote, noteUpdate, noteUpdateTitles);
 							
 							// Suspend this thread while the GUI is presented to
 							// the user.
@@ -311,7 +323,7 @@ Logger.Debug ("8");
 				SetState (SyncState.Downloading);
 
 			// TODO: Figure out why GUI doesn't always update smoothly
-
+			
 			// Process updates from the server; the bread and butter of sync!
 			foreach (NoteUpdate noteUpdate in noteUpdates.Values) {
 				Note existingNote = FindNoteByUUID (noteUpdate.UUID);
@@ -327,7 +339,7 @@ Logger.Debug ("8");
 				        noteUpdate.Title));
 					// Note already exists locally, but has been modified since last sync; prompt user
 					if (NoteConflictDetected != null) {
-						NoteConflictDetected (NoteMgr, existingNote, noteUpdate);
+						NoteConflictDetected (NoteMgr, existingNote, noteUpdate, noteUpdateTitles);
 						
 						// Suspend this thread while the GUI is presented to
 						// the user.
@@ -343,8 +355,14 @@ Logger.Debug ("8");
 				}
 			}
 
+			// Note deletion may affect the GUI, so we have to use the
+			// delegate to run in the main gtk thread.
+			// To be consistent, any exceptions in the delgate will be caught
+			// and then rethrown in the synchronization thread.
+			Exception mainThreadException = null;
 			AutoResetEvent evt = new AutoResetEvent (false);
 			Gtk.Application.Invoke (delegate {
+				try {
 				// Make list of all local notes
 				List<Note> localNotes = new List<Note> ();
 				foreach (Note note in NoteMgr.Notes)
@@ -363,10 +381,15 @@ Logger.Debug ("8");
 						NoteMgr.Delete (note);
 					}
 				}
-			evt.Set ();
+				evt.Set ();
+				} catch (Exception e) {
+					mainThreadException = e;
+				}
 			});
 			
 			evt.WaitOne ();
+			if (mainThreadException != null)
+				throw mainThreadException;
 			
 			// TODO: Add following updates to syncDialog treeview
 			
@@ -377,6 +400,7 @@ Logger.Debug ("8");
 			foreach (Note note in NoteMgr.Notes) {
 				if (client.GetRevision (note) == -1) {
 					// This is a new note that has never been synchronized to the server
+					// TODO: *OR* this is a note that we lost revision info for!!!
 					note.Save ();
 					newOrModifiedNotes.Add (note);
 					if (NoteSynchronized != null)
@@ -436,8 +460,26 @@ Logger.Debug ("8");
 			Logger.Debug ("Sync: New revision: {0}", client.LastSynchronizedRevision);
 			
 			SetState (SyncState.Idle);
-			syncThread = null;
-			addin.PostSyncCleanup (); // TODO: try/finally this?
+			
+			} catch (Exception e) { // top-level try
+				Logger.Error ("Synchronization failed with the following exception: " +
+					      e.Message + "\n" +
+					      e.StackTrace);
+				// TODO: Report graphically to user
+				try {
+					SetState (SyncState.Idle); // stop progress
+					SetState (SyncState.Failed);
+				} catch {}
+			} finally {
+				syncThread = null;
+				try {
+					addin.PostSyncCleanup ();
+				} catch (Exception e) {
+					Logger.Error ("Error cleaning up addin after sync: " +
+						      e.Message + "\n" +
+						      e.StackTrace);
+				}
+			}
 		}
 		
 		/// <summary>
@@ -455,33 +497,57 @@ Logger.Debug ("8");
 			
 		private static void CreateNoteInMainThread (NoteUpdate noteUpdate)
 		{
+			// Note creation may affect the GUI, so we have to use the
+			// delegate to run in the main gtk thread.
+			// To be consistent, any exceptions in the delgate will be caught
+			// and then rethrown in the synchronization thread.
+			Exception mainThreadException = null;
 			AutoResetEvent evt = new AutoResetEvent (false);
 			Gtk.Application.Invoke (delegate {
-				Note existingNote = NoteMgr.CreateWithGuid (noteUpdate.Title, noteUpdate.UUID);
-				UpdateLocalNote (existingNote, noteUpdate, NoteSyncType.DownloadNew);
+				try {
+					Note existingNote = NoteMgr.CreateWithGuid (noteUpdate.Title, noteUpdate.UUID);
+					UpdateLocalNote (existingNote, noteUpdate, NoteSyncType.DownloadNew);
+				} catch (Exception e) {
+					mainThreadException = e;
+				}
 						
 				evt.Set ();
 			});
 			
 			evt.WaitOne ();
+			if (mainThreadException != null)
+				throw mainThreadException;
 		}
 		
 		private static void UpdateNoteInMainThread (Note existingNote, NoteUpdate noteUpdate)
 		{
+			// Note update may affect the GUI, so we have to use the
+			// delegate to run in the main gtk thread.
+			// To be consistent, any exceptions in the delgate will be caught
+			// and then rethrown in the synchronization thread.
+			Exception mainThreadException = null;
 			AutoResetEvent evt = new AutoResetEvent (false);
 			Gtk.Application.Invoke (delegate {
-				UpdateLocalNote (existingNote, noteUpdate, NoteSyncType.DownloadModified);
-
+				try {
+					UpdateLocalNote (existingNote, noteUpdate, NoteSyncType.DownloadModified);
+				} catch (Exception e) {
+					mainThreadException = e;
+				}
+						
 				evt.Set ();
 			});
 			
-			evt.WaitOne ();				
+			evt.WaitOne ();
+			if (mainThreadException != null)
+				throw mainThreadException;			
 		}
 		
 		private static void UpdateLocalNote (Note localNote, NoteUpdate serverNote, NoteSyncType syncType)
 		{
 			// In each case, update existingNote's content and revision
-			localNote.LoadForeignNoteXml (serverNote.XmlContent);
+			try {
+				localNote.LoadForeignNoteXml (serverNote.XmlContent);
+			} catch {} // TODO: Handle exception in case that serverNote.XmlContent is invalid XML
 			client.SetRevision (localNote, serverNote.LatestRevision);
 
 			// Update dialog's sync status
